/*
Copyright 2004 John Tsiombikas <nuclear@siggraph.org>

This file is part of the X-Ray ray-tracing library.

X-Ray is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

X-Ray is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with X-Ray; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string>
#include <stack>
#include <cstdio>
#include <cstring>
#include <expat.h>
#include "loader.hpp"
#include "mat_phong.hpp"
#include "texman.hpp"
#include "common/err_msg.h"

#define BUF_SIZE	4096

using namespace xray;
using std::stack;
using std::string;

// ------------- XML tag handlers ----------------
static void StartTag(void *user_data, const char *name, const char **atts);
static void EndTag(void *user_data, const char *name);

static void CameraHandler(void *user_data, const char *str, int len);
static void LightHandler(void *user_data, const char *str, int len);

// object handlers
static void SphereHandler(void *user_data, const char *str, int len);
static void PlaneHandler(void *user_data, const char *str, int len);

// material handlers
static void MatPhongHandler(void *user_data, const char *str, int len);
// -----------------------------------------------

static const char *GetAttr(const char *name, const char **atts);
static Vector3 GetVec3(const char *str);
static Color GetColor(const char *str);
char *MakeCStr(const char *str, int len);

static XML_Parser xml;

static XFormNode *node;
static Material *mat;

static void (*ObjHandler)(void*, const char*, int);

static int level;
static stack<string> tag_stack;


Scene *xray::LoadXML(const char *fname) {
	FILE *fp;

	if(!(fp = fopen(fname, "r"))) {
		return 0;
	}
	
	Scene *scene = new Scene;

	xml = XML_ParserCreate(0);
	XML_SetElementHandler(xml, StartTag, EndTag);
	XML_SetUserData(xml, scene);

	char *buf = new char[BUF_SIZE];
	
	bool done = false;
	while(!done) {
		size_t bytes = fread(buf, 1, BUF_SIZE, fp);
		done = bytes < BUF_SIZE;

		XML_Parse(xml, buf, bytes, done);
	}

	delete [] buf;
	XML_ParserFree(xml);
	
	return scene;
}


static void StartTag(void *user_data, const char *name, const char **atts) {

	string sname(name);
	
	tag_stack.push(sname);
	level++;
	
	if(sname == "camera") {
		const char *type;
		if((type = GetAttr("type", atts)) && string(type) == "target") {
			node = new TargetCamera;
		} else {
			node = new Camera;
		}
		XML_SetCharacterDataHandler(xml, CameraHandler);
		
	} else if(sname == "light") {
		const char *type = GetAttr("type", atts);
		if(type && string(type) == "point") {
			node = new PointLight;
			XML_SetCharacterDataHandler(xml, LightHandler);
			
		} else if(type && string(type) == "area") {
			node = new AreaLight;

			const char *shadow_samples = GetAttr("shadow-samples", atts);
			if(shadow_samples) {
				dynamic_cast<AreaLight*>(node)->SetShadowSamples(atoi(shadow_samples));
			}

			XML_SetCharacterDataHandler(xml, LightHandler);
		}
		
	} else if(sname == "object") {
		const char *type;
		if(!(type = GetAttr("type", atts))) {
			warning("invalid object tag, no type specified");
			return;
		}

		if(string(type) == "sphere") {
			node = new QuadrObject;
			((QuadrObject*)node)->SetSurface(new Sphere);
			ObjHandler = SphereHandler;
		} else if(string(type) == "plane") {
			node = new QuadrObject;
			((QuadrObject*)node)->SetSurface(new Plane);
			ObjHandler = PlaneHandler;
		} else {
			error("only spheres and planes available at the moment");
		}
		
		XML_SetCharacterDataHandler(xml, ObjHandler);
		
	} else if(sname == "material") {
		const char *type;
		if(!(type = GetAttr("type", atts))) {
			warning("invalid material tag, no type specified");
			return;
		}

		if(string(type) == "phong") {
			mat = new MatPhong;
			XML_SetCharacterDataHandler(xml, MatPhongHandler);
		} else {
			error("only phong material available at the moment");
		}
	}
			
}

static void EndTag(void *user_data, const char *name) {
	Scene *scene = (Scene*)user_data;
	string sname = string(name);

	if(sname != tag_stack.top()) {
		error("panic!!!! internal tag stack corrupted?");
		exit(-1);
	}
	
	tag_stack.pop();
	level--;

	if(sname == "camera") {
		if(!node) return;
		Camera *cam = (Camera*)node;
		info("adding camera: %s", cam->GetName().c_str());
		scene->AddCamera(cam);
		node = 0;
		XML_SetCharacterDataHandler(xml, 0);

	} else if(sname == "light") {
		if(!node) return;
		Light *light = (Light*)node;
		info("adding light: %s", light->GetName().c_str());
		scene->AddLight(light);
		node = 0;
		XML_SetCharacterDataHandler(xml, 0);
		
	} else if(sname == "object") {
		if(!node) return;
		Object *obj = (Object*)node;
		if(mat) {
			obj->SetMaterial(mat);
			mat = 0;
		}

		info("adding object: %s", obj->GetName().c_str());
		scene->AddObject(obj);
		node = 0;
		ObjHandler = 0;
		XML_SetCharacterDataHandler(xml, 0);
		
	} else if(sname == "material") {
		XML_SetCharacterDataHandler(xml, ObjHandler);
	}

}

static void CameraHandler(void *user_data, const char *str, int len) {
	char *cstr = MakeCStr(str, len);
	
	string tag = tag_stack.top();
	Camera *cam = (Camera*)node;
	TargetCamera *tcam = dynamic_cast<TargetCamera*>(cam);
	
	if(tag == "name") {
		cam->SetName(cstr);
	} else if(tag == "position") {
		cam->SetPosition(GetVec3(cstr));
	} else if(tag == "fov") {
		cam->SetFOV(DEG_TO_RAD(atof(cstr)));
	} else if(tag == "target") {
		if(tcam) {
			tcam->SetTarget(GetVec3(cstr));
		} else {
			warning("camera target specified on free camera, ignoring");
		}
	}

	delete [] cstr;
}

static void LightHandler(void *user_data, const char *str, int len) {
	char *cstr = MakeCStr(str, len);
	
	string tag = tag_stack.top();
	Light *light = (PointLight*)node;
	AreaLight *alight = dynamic_cast<AreaLight*>(light);
	
	if(tag == "name") {
		light->SetName(cstr);
	} else if(tag == "position") {
		light->SetPosition(GetVec3(cstr));
	} else if(tag == "color") {
		light->SetColor(GetColor(cstr));
	} else if(tag == "intensity") {
		light->SetIntensity(atof(cstr));
	} else if(tag == "radius") {
		if(alight) {
			alight->SetRadius(atof(cstr));
		} else {
			warning("radius tag encountered in a non-area-light, ignoring");
		}
	}
			

	delete [] cstr;
}

static void SphereHandler(void *user_data, const char *str, int len) {
	char *cstr = MakeCStr(str, len);

	string tag = tag_stack.top();
	QuadrObject *sph = (QuadrObject*)node;
	Sphere *surf = (Sphere*)sph->GetSurface();

	if(tag == "name") {
		sph->SetName(cstr);
	} else if(tag == "two-sided") {
		if(!strcmp(cstr, "false")) {
			sph->SetTwoSided(false);
		} else if(!strcmp(cstr, "true")) {
			sph->SetTwoSided(true);
		} else {
			warning("unrecognized value in <two-sided> tag");
		}
	} else if(tag == "center") {
		Vector3 pos = GetVec3(cstr);
		//TODO: remove quick hack and merge with the XFormNode interface
		sph->SetPosition(pos);
		surf->SetPosition(pos);
	} else if(tag == "radius") {
		surf->SetRadius(atof(cstr));
	}

	delete [] cstr;
}

static void PlaneHandler(void *user_data, const char *str, int len) {
	char *cstr = MakeCStr(str, len);

	string tag = tag_stack.top();
	QuadrObject *plane = (QuadrObject*)node;
	Plane *surf = (Plane*)plane->GetSurface();

	if(tag == "name") {
		plane->SetName(cstr);
	} else if(tag == "two-sided") {
		if(!strcmp(cstr, "false")) {
			plane->SetTwoSided(false);
		} else if(!strcmp(cstr, "true")) {
			plane->SetTwoSided(true);
		} else {
			warning("unrecognized value in <two-sided> tag");
		}
	} else if(tag == "normal") {
		Vector3 normal = GetVec3(cstr);
		//TODO: same as above, XFormNode etc...
		surf->SetNormal(normal);
	} else if(tag == "position") {
		Vector3 pos = GetVec3(cstr);
		plane->SetPosition(pos);
		surf->SetPosition(pos);
	}

	delete [] cstr;
}

static void MatPhongHandler(void *user_data, const char *str, int len) {
	char *cstr = MakeCStr(str, len);

	string tag = tag_stack.top();
	tag_stack.pop();
	string ptag = tag_stack.top();
	tag_stack.push(tag);

	if(ptag == "ior") {
		MatAttr *attr = mat->GetAttrPtr(MAT_IOR);
		if(tag == "val") attr->SetValue(atof(cstr)); else attr->SetMap(GetTexture(cstr));
	} else if(ptag == "power") {
		MatAttr *attr = mat->GetAttrPtr(MAT_PHONG_POWER);
		if(tag == "val") attr->SetValue(atof(cstr)); else attr->SetMap(GetTexture(cstr));
	} else if(ptag == "ambient") {
		MatAttr *attr = mat->GetAttrPtr(MAT_PHONG_AMBIENT);
		if(tag == "val") attr->SetVector(GetVec3(cstr)); else attr->SetMap(GetTexture(cstr));
	} else if(ptag == "diffuse") {
		MatAttr *attr = mat->GetAttrPtr(MAT_PHONG_DIFFUSE);
		if(tag == "val") attr->SetVector(GetVec3(cstr)); else attr->SetMap(GetTexture(cstr));
	} else if(ptag == "specular") {
		MatAttr *attr = mat->GetAttrPtr(MAT_PHONG_SPECULAR);
		if(tag == "val") attr->SetVector(GetVec3(cstr)); else attr->SetMap(GetTexture(cstr));
	} else if(ptag == "emissive") {
		MatAttr *attr = mat->GetAttrPtr(MAT_PHONG_EMISSIVE);
		if(tag == "val") attr->SetVector(GetVec3(cstr)); else attr->SetMap(GetTexture(cstr));
	} else if(ptag == "reflection") {
		MatAttr *attr = mat->GetAttrPtr(MAT_REFLECT);
		if(tag == "val") attr->SetValue(atof(cstr)); else attr->SetMap(GetTexture(cstr));
	} else if(ptag == "transmission") {
		MatAttr *attr = mat->GetAttrPtr(MAT_TRANSMIT);
		if(tag == "val") attr->SetValue(atof(cstr)); else attr->SetMap(GetTexture(cstr));
	} else if(ptag == "glossiness") {
		MatAttr *attr = mat->GetAttrPtr(MAT_GLOSSINESS);
		if(tag == "val") attr->SetValue(atof(cstr)); else attr->SetMap(GetTexture(cstr));
	}

	delete [] cstr;
}

// --- helpers ---

static const char *GetAttr(const char *name, const char **atts) {
	int i = 0;
	while(atts[i]) {
		if(!strcmp(name, atts[i])) {
			return atts[i+1];
		}
		i += 2;
	}
	return 0;
}

static Vector3 GetVec3(const char *str) {
	Vector3 v;

	if(!str) {
		warning("bad input on GetVec3()");
		return v;
	}
	v.x = atof(str);
	
	if(!(str = strchr(str, ','))) {
		warning("bad input on GetVec3()");
		return v;
	}
	v.y = atof(++str);
	
	if(!(str = strchr(str, ','))) {
		warning("bad input on GetVec3()");
		return v;
	}
	v.z = atof(++str);

	return v;
}

static Color GetColor(const char *str) {
	Vector3 v = GetVec3(str);
	return Color(v.x, v.y, v.z);
}

char *MakeCStr(const char *str, int len) {
	char *cstr = new char[len+1];
	memcpy(cstr, str, len);
	cstr[len] = 0;
	return cstr;
}
