#include <limits>
#include "raytracer.hpp"
#include "material.hpp"
#include "mat_phong.hpp"

#include "gfx/base_cam.hpp"
#include "common/err_msg.h"
#include "common/config_parser.h"

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

using namespace xray;

static XRayContext *active_xrc;

// finds the configuration file by searching at ./xrayrc $(HOME)/.xrayrc and /etc/xrayrc
static const char *FindConfigFile() {
	FILE *fp;
	static char fname[512];
	
	strcpy(fname, "xrayrc");
	if((fp = fopen(fname, "r"))) {
		fclose(fp);
		return fname;
	}
	
	char *home_dir = getenv("HOME");
	if(home_dir) {
		strcpy(fname, home_dir);
		strcat(fname, "/.xrayrc");

		if((fp = fopen(fname, "r"))) {
			fclose(fp);
			return fname;
		}
	}

	strcpy(fname, "/etc/xrayrc");
	if((fp = fopen(fname, "r"))) {
		fclose(fp);
		return fname;
	}

	
	return 0;
}

XRayContext *xray::Init(int x, int y) {
	info("Initializing X-Ray visualization system");
	info("setting rendering resolution: %dx%d", x, y);

	XRayContext *xray = new XRayContext;
	xray->x = x;
	xray->y = y;
	xray->pbuf = new PixelBuffer(x, y);
	xray->pix_pos = new Vector3[x * y];
	
	xray->rays_per_pixel = 1;
	xray->max_ray_depth = 5;
	xray->ray_energy_threshold = 0.01;
	xray->area_light_samples = 20;
	xray->glossy_samples = 20;
	xray->max_diffuse_bounces = 0;
	xray->scene = 0;
	xray->ProgressCallback = 0;

	const char *cfg_file = FindConfigFile();
	if(cfg_file) {
		SetParserState(PS_CommentSymbol, '#');
		
		if(LoadConfigFile(cfg_file) == -1) {
			warning("could not read config file...");
		}

		const ConfigOption *cfg_opt;
		while((cfg_opt = GetNextOption())) {
			if(!strcmp(cfg_opt->option, "rays_per_pixel")) {
				xray->rays_per_pixel = cfg_opt->int_value;
				info("(config) rays per pixel: %d", cfg_opt->int_value);
				
			} else if(!strcmp(cfg_opt->option, "max_ray_depth")) {
				xray->max_ray_depth = cfg_opt->int_value;
				info("(config) max ray recursion depth: %d", cfg_opt->int_value);

			} else if(!strcmp(cfg_opt->option, "ray_energy_threshold")) {
				xray->ray_energy_threshold = cfg_opt->flt_value;
				info("(config) ray energy threshold: %f", cfg_opt->flt_value);
				
			} else if(!strcmp(cfg_opt->option, "area_light_samples")) {
				xray->area_light_samples = cfg_opt->int_value;
				info("(config) area light shadow samples: %d", cfg_opt->int_value);
				
			} else if(!strcmp(cfg_opt->option, "glossy_samples")) {
				xray->glossy_samples = cfg_opt->int_value;
				info("(config) glossy samples: %d", cfg_opt->int_value);

			} else if(!strcmp(cfg_opt->option, "max_diffuse_bounces")) {
				xray->max_diffuse_bounces = cfg_opt->int_value;
				info("(config) max diffuse bounces: %d", cfg_opt->int_value);
				
			} else {
				warning("(config) unrecognized option: %s", cfg_opt->option);
			}
		}
	} else {
		warning("config file not found, using defaults");
	}

	if(!active_xrc) SetActiveXRayContext(xray);
	ResetRenderStats();

	return xray;
}

void xray::Destroy(XRayContext *xrc) {
	if(!xrc) xrc = active_xrc;

	if(xrc) {
		delete xrc->pbuf;
		delete [] xrc->pix_pos;
		delete xrc;
	}
}

void xray::SetActiveXRayContext(XRayContext *xrc) {
	active_xrc = xrc;
}

XRayContext *xray::GetActiveXRayContext() {
	return active_xrc;
}

void xray::SetProgressCallback(void (*ProgressCallback)(float)) {
	active_xrc->ProgressCallback = ProgressCallback;
	if(active_xrc->scene) active_xrc->scene->SetProgressCallback(ProgressCallback);
}

const PixelBuffer *xray::GetPixelBuffer() {
	return active_xrc->pbuf;
}

void xray::SetActiveScene(Scene *scene) {
	active_xrc->scene = scene;
	scene->SetProgressCallback(active_xrc->ProgressCallback);
}

Scene *xray::GetActiveScene() {
	return active_xrc->scene;
}

void xray::ResetRenderStats() {
	if(!active_xrc) return;
	memset(&active_xrc->stats, 0, sizeof(RenderStats));
}

RenderStats *xray::GetRenderStats() {
	return &active_xrc->stats;
}

/**
 * recursive ray trace algorithm.
 * (the call to Shade() may result in more invocations of ShootRay)
 */
Color xray::ShootRay(const Ray &ray, int ray_depth) {
	Scene *scene = active_xrc->scene;

	if(ray_depth == active_xrc->max_ray_depth) {
		return scene->background;
	}

	// statistics
	if(ray_depth == 0) {
		active_xrc->stats.primary_rays++;
	} else {
		active_xrc->stats.secondary_rays++;
	}

	SurfPoint nearest_point;
	nearest_point.t = std::numeric_limits<scalar_t>::max();
	Object *hit_obj = 0;

	int obj_count = scene->objects.size();
	for(int i=0; i<obj_count; i++) {
		Object *obj = scene->objects[i];

		SurfPoint point;
		if(obj->FindIntersection(ray, &point)) {
			if(point.t < nearest_point.t) {
				nearest_point = point;
				hit_obj = obj;
			}
		}
	}

	if(!hit_obj) return scene->background;

	if(active_xrc->max_diffuse_bounces > 0) {
		return ShadeMonteCarlo(ray, nearest_point, hit_obj, ray_depth);
	}
	return Shade(ray, nearest_point, hit_obj, ray_depth);
}

/** shadow ray casting */
scalar_t xray::ShadowRay(const Vector3 &pt, const Light *light) {
	Scene *scene = active_xrc->scene;
	scalar_t result = 1.0;

	Vector3 lpos = light->GetPRS(scene->anim.current_time).position;
	const AreaLight *alight = dynamic_cast<const AreaLight*>(light);

	if(alight) {	// calculate area-light shadows
		scalar_t rad = alight->GetRadius();
		scalar_t two_rad = rad * 2.0;
		
		int area_light_samples = active_xrc->area_light_samples;
		int tmp = alight->GetShadowSamples();
		if(tmp != -1) {
			area_light_samples = tmp;
		}
		
		Ray *rays = new Ray[area_light_samples];
		for(int i=0; i<area_light_samples; i++) {
			Vector3 ray_end = lpos + Vector3(frand(two_rad)-rad, frand(two_rad)-rad, frand(two_rad)-rad);
			rays[i] = Ray(pt, ray_end - pt);
		}

		int rays_in_light = area_light_samples;
		
		// for every sample ray
		for(int j=0; j<area_light_samples; j++) {
			
			// check every object
			int obj_count = scene->objects.size();
			for(int i=0; i<obj_count; i++) {
				Object *obj = scene->objects[i];

				SurfPoint point;
				if(obj->FindIntersection(rays[j], &point) && point.t < 1.0) {
					Material *mat = obj->GetMaterial();
					scalar_t trans = mat->GetAttr(MAT_TRANSMIT).GetValue();
					if(trans < small_number) {
						rays_in_light--;
					}
				}
			}
		}

		delete [] rays;

		result = (scalar_t)rays_in_light / (scalar_t)area_light_samples;

		active_xrc->stats.shadow_rays += area_light_samples;

	} else {	// point-light shadows
		Ray ray(pt, lpos - pt);

		int obj_count = scene->objects.size();
		for(int i=0; i<obj_count; i++) {
			Object *obj = scene->objects[i];

			SurfPoint point;
			if(obj->FindIntersection(ray, &point) && point.t < 1.0) {
				Material *mat = obj->GetMaterial();
				scalar_t trans = mat->GetAttr(MAT_TRANSMIT).GetValue();
				result *= trans;
			}
		}

		active_xrc->stats.shadow_rays++;
	}

	return result;
}

Color xray::Shade(const Ray &ray, SurfPoint pt, const Object *obj, int ray_depth) {
	Scene *scene = active_xrc->scene;
	bool entering = true;
	
	if(DotProduct(ray.dir, pt.normal) > 0.0) {
		pt.normal = -pt.normal;
		entering = false;
	}
	
	TexCoord tc = obj->GetTexCoords(pt.pos);
	Vector3 vref = ray.dir.Normalized().Reflection(pt.normal);

	MatPhong *mat;
	if(!(mat = dynamic_cast<MatPhong*>(obj->GetMaterial()))) {
		error("only phong materials supported currently\n");
		exit(0);
	}

	Color diffuse = 0.0;
	Color specular = 0.0;

	int light_count = scene->lights.size();
	for(int i=0; i<light_count; i++) {
		Light *light = scene->lights[i];
		
		Vector3 lpos = light->GetPRS(scene->anim.current_time).position;
		Vector3 ldir = lpos - pt.pos;
		//scalar_t light_dist = ldir.Length();
		ldir.Normalize();
				
		// cast shadow ray
		scalar_t shadow_factor = ShadowRay(pt.pos, light);
		if(shadow_factor == 0.0) continue;

		Color lcol = light->GetColor() * light->GetIntensity();


		// calculate diffuse contribution
		scalar_t diffuse_coef = std::max(0.0, DotProduct(ldir, pt.normal));
		diffuse += Color(mat->GetAttr(MAT_PHONG_DIFFUSE)(tc)) * lcol * diffuse_coef * shadow_factor;

		// calculate specular contribution
		scalar_t spec_pow = mat->GetAttr(MAT_PHONG_POWER)(tc).x;
		scalar_t specular_coef = pow(std::max(0.0, DotProduct(ldir, vref)), spec_pow);
		specular += Color(mat->GetAttr(MAT_PHONG_SPECULAR)(tc)) * lcol * specular_coef * shadow_factor;
	}
	
	scalar_t refl_coeff = mat->GetAttr(MAT_REFLECT)(tc).x;

	scalar_t mat_ior = mat->GetAttr(MAT_IOR)(tc).x;
	if(mat_ior > 1.0) {		// dielectric
		refl_coeff *= std::max(0.0, std::min(Fresnel(DotProduct(-ray.dir, pt.normal), mat_ior), 1.0));
	}

	Color refl;
	if(refl_coeff > active_xrc->ray_energy_threshold) {
		Ray refl_ray = ray;
		refl_ray.origin = pt.pos;
		refl_ray.dir = ray.dir.Reflection(pt.normal);
		refl_ray.energy *= refl_coeff;
		
		float glossiness = mat->GetAttr(MAT_GLOSSINESS)(tc).x;
		if(glossiness < small_number) {
			refl = ShootRay(refl_ray, ray_depth + 1);
		} else {
			int samples = std::max(1, (active_xrc->glossy_samples / (ray_depth + 1)) / active_xrc->rays_per_pixel);

			refl = SampleSolidAngle(refl_ray, glossiness, samples, ray_depth + 1);
			active_xrc->stats.gloss_rays += samples;
		}
	}
	
	scalar_t trans_coeff = mat->GetAttr(MAT_TRANSMIT)(tc).x;

	Color trans;
	if(trans_coeff > active_xrc->ray_energy_threshold) {
		Ray trans_ray = ray;
		trans_ray.origin = pt.pos;
		trans_ray.energy *= trans_coeff;
		
		scalar_t new_ior;
		if(entering) {
			new_ior = mat_ior;
			trans_ray.Enter(new_ior);
		} else {
			trans_ray.Leave();
			new_ior = trans_ray.ior;
		}
		trans_ray.dir = ray.dir.Refraction(pt.normal, ray.ior, new_ior);

		if(DotProduct(trans_ray.dir, pt.normal) > 0.0) {	// if TIR occured
			if(entering) {
				trans_ray.Leave();	// we did not actually enter...
			} else {
				trans_ray.Enter(mat_ior);	// we did not actually leave...
			}
		}

		trans = ShootRay(trans_ray, ray_depth + 1);
	}

	Color ambient = Color(mat->GetAttr(MAT_PHONG_AMBIENT)(tc)) * scene->ambient;
	Color emissive = Color(mat->GetAttr(MAT_PHONG_EMISSIVE)(tc));
	
	return ambient + emissive + diffuse + specular + refl * refl_coeff + trans * trans_coeff;
}


Color xray::ShadeMonteCarlo(const Ray &ray, SurfPoint pt, const Object *obj, int ray_depth) {
	Scene *scene = active_xrc->scene;
	bool entering = true;
	
	if(DotProduct(ray.dir, pt.normal) > 0.0) {
		pt.normal = -pt.normal;
		entering = false;
	}
	
	TexCoord tc = obj->GetTexCoords(pt.pos);
	Vector3 vref = ray.dir.Normalized().Reflection(pt.normal);

	MatPhong *mat;
	if(!(mat = dynamic_cast<MatPhong*>(obj->GetMaterial()))) {
		error("only phong materials supported currently\n");
		exit(0);
	}

	Color diffuse = 0.0;
	Color specular = 0.0;
	
	Color mat_diffuse = mat->GetAttr(MAT_PHONG_DIFFUSE)(tc);

	// ---------- calculate direct illumination ---------
	int light_count = scene->lights.size();
	for(int i=0; i<light_count; i++) {
		Light *light = scene->lights[i];
		
		Vector3 lpos = light->GetPRS(scene->anim.current_time).position;
		Vector3 ldir = lpos - pt.pos;
		//scalar_t light_dist = ldir.Length();
		ldir.Normalize();
				
		// cast shadow ray
		scalar_t shadow_factor = ShadowRay(pt.pos, light);
		if(shadow_factor == 0.0) continue;

		Color lcol = light->GetColor() * light->GetIntensity();


		// calculate diffuse contribution
		scalar_t diffuse_coef = std::max(0.0, DotProduct(ldir, pt.normal));
		diffuse += mat_diffuse * lcol * diffuse_coef * shadow_factor;

		// calculate specular contribution
		scalar_t spec_pow = mat->GetAttr(MAT_PHONG_POWER)(tc).x;
		scalar_t specular_coef = pow(std::max(0.0, DotProduct(ldir, vref)), spec_pow);
		specular += Color(mat->GetAttr(MAT_PHONG_SPECULAR)(tc)) * lcol * specular_coef * shadow_factor;
	}

	// ------ caclulate indirect diffuse --------
	scalar_t mat_diffuse_value = std::max(std::max(mat_diffuse.r, mat_diffuse.g), mat_diffuse.b);
	if(mat_diffuse_value > active_xrc->ray_energy_threshold) {
		// create a random reflection vector within the hemisphere
		Vector3 rnd_dir = pt.normal + Vector3(frand(pi) - half_pi, frand(pi) - half_pi, frand(pi) - half_pi);
		
		Ray diffuse_ray = ray;
		diffuse_ray.dir = rnd_dir;

		diffuse += ShootRay(diffuse_ray, ray_depth + 1) * mat_diffuse;
	}
	
	// ----- calculate indirect specular -------
	scalar_t refl_coeff = mat->GetAttr(MAT_REFLECT)(tc).x;

	scalar_t mat_ior = mat->GetAttr(MAT_IOR)(tc).x;
	if(mat_ior > 1.0) {		// dielectric
		refl_coeff *= std::max(0.0, std::min(Fresnel(DotProduct(-ray.dir, pt.normal), mat_ior), 1.0));
	}

	Color refl;
	if(refl_coeff > active_xrc->ray_energy_threshold) {
		Ray refl_ray = ray;
		refl_ray.origin = pt.pos;
		refl_ray.dir = ray.dir.Reflection(pt.normal);
		refl_ray.energy *= refl_coeff;
		
		float angle = mat->GetAttr(MAT_GLOSSINESS)(tc).x;
		if(angle < small_number) {
			refl = ShootRay(refl_ray, ray_depth + 1);
		} else {
			Matrix4x4 rot;
			scalar_t half_angle = angle / 2.0;
			rot.SetRotation(Vector3(frand(angle) - half_angle, frand(angle) - half_angle, frand(angle) - half_angle));
		
			refl_ray.dir.Transform(rot);
			refl = ShootRay(refl_ray, ray_depth + 1);
		}
	}

	scalar_t trans_coeff = mat->GetAttr(MAT_TRANSMIT)(tc).x;

	Color trans;
	if(trans_coeff > active_xrc->ray_energy_threshold) {
		Ray trans_ray = ray;
		trans_ray.origin = pt.pos;
		trans_ray.energy *= trans_coeff;
		
		scalar_t new_ior;
		if(entering) {
			new_ior = mat_ior;
			trans_ray.Enter(new_ior);
		} else {
			trans_ray.Leave();
			new_ior = trans_ray.ior;
		}
		trans_ray.dir = ray.dir.Refraction(pt.normal, ray.ior, new_ior);

		if(DotProduct(trans_ray.dir, pt.normal) > 0.0) {	// if TIR occured
			if(entering) {
				trans_ray.Leave();	// we did not actually enter...
			} else {
				trans_ray.Enter(mat_ior);	// we did not actually leave...
			}
		}

		trans = ShootRay(trans_ray, ray_depth + 1);
	}

	Color ambient = Color(mat->GetAttr(MAT_PHONG_AMBIENT)(tc)) * scene->ambient;
	Color emissive = Color(mat->GetAttr(MAT_PHONG_EMISSIVE)(tc));
	
	return ambient + emissive + diffuse + specular + refl * refl_coeff + trans * trans_coeff;

}


Color xray::SampleSolidAngle(const Ray &ray, scalar_t angle, int samples, int ray_depth) {
	Vector3 accumulator;

	for(int i=0; i<samples; i++) {
		Matrix3x3 rot;
		scalar_t half_angle = angle / 2.0;
		rot.SetRotation(Vector3(frand(angle) - half_angle, frand(angle) - half_angle, frand(angle) - half_angle));
		
		Ray sample_ray = ray;
		sample_ray.dir.Transform(rot);

		Color tmp = ShootRay(sample_ray, ray_depth);
		accumulator.x += tmp.r;
		accumulator.y += tmp.g;
		accumulator.z += tmp.b;
	}

	return Color(accumulator / samples);
}

scalar_t xray::Fresnel(scalar_t cos_theta, scalar_t ior) {
	scalar_t g = sqrt(SQ(ior) + SQ(cos_theta) - 1.0);

	scalar_t g_plus_c = g + cos_theta;
	scalar_t g_min_c = g - cos_theta;

	return 0.5 * (SQ(g_min_c) / SQ(g_plus_c)) * (1.0 + SQ(cos_theta * g_plus_c - 1) / SQ(cos_theta * g_min_c - 1));
}
