/*
Copyright 2004 John Tsiombikas <nuclear@siggraph.org>

This file is part of the X-Ray ray-tracing library.

X-Ray is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

X-Ray is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with X-Ray; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string>
#include <algorithm>
#include "raytracer.hpp"
#include "scene.hpp"
#include "common/err_msg.h"
#include "common/timer.h"

using namespace xray;

/** scene constructor */
Scene::Scene() {
	background = 0.0;
	ambient = 0.0;
	
	manage_elements = true;

	active_cam = 0;
	ProgressCallback = 0;
	
	if(!GetActiveScene()) SetActiveScene(this);
}

/** 
 * scene destructor.
 *
 * if manage_elements is true, it deletes all data in the trees,
 * otherwise it doesn't touch the data (take care to deallocate
 * them properly yourself).
 */
Scene::~Scene() {
	if(manage_elements) {
		for(int i=0; i<(int)objects.size(); i++) delete objects[i];
		for(int i=0; i<(int)lights.size(); i++) delete lights[i];
		for(int i=0; i<(int)cameras.size(); i++) delete cameras[i];
		for(int i=0; i<(int)curves.size(); i++) delete curves[i];
	}
}

void Scene::CalcPixelPositions() {
	XRayContext *xrc = GetActiveXRayContext();

	scalar_t aspect = (scalar_t)xrc->x / (scalar_t)xrc->y;

	scalar_t xsize = 2.0;
	scalar_t ysize = 2.0 / aspect;

	scalar_t xpos = -xsize / 2.0;
	scalar_t ypos = ysize / 2.0;

	scalar_t dx = xsize / xrc->x;
	scalar_t dy = -ysize / xrc->y;

	scalar_t dist = 1.0 / tan(active_cam->GetFOV()/2.0);

	Vector3 *pptr = xrc->pix_pos;
	for(int y=0; y<xrc->y; y++) {
		for(int x=0; x<xrc->x; x++) {
			*pptr++ = Vector3(xpos, ypos, dist).Transformed(cam_xform);
			xpos += dx;
		}
		xpos = -xsize / 2.0;
		ypos += dy;
	}

	xrc->eye_pos = active_cam->GetPRS(anim.current_time).position;
}

void Scene::SetProgressCallback(void (*ProgressCallback)(float)) {
	this->ProgressCallback = ProgressCallback;
}

/** adds an object to the scene */
void Scene::AddObject(Object *obj) {
	objects.push_back(obj);
}

/** adds a light to the scene */
void Scene::AddLight(Light *light) {
	lights.push_back(light);
}

/** adds a camera to the scene */
void Scene::AddCamera(Camera *cam) {
	cameras.push_back(cam);
	if(!active_cam) SetActiveCamera(cam);
}

/** adds a curve to the scene */
void Scene::AddCurve(Curve *curve) {
	curves.push_back(curve);
}


/**
 * removes an object from the scene. 
 * if the second argument (del) is true, it deletes it as well.
 */
void Scene::RemoveObject(Object *obj, bool del) {
	std::vector<Object*>::iterator iter = std::find(objects.begin(), objects.end(), obj);
	
	if(iter != objects.end()) {
		objects.erase(iter);
		if(del) delete obj;
	}
}

/**
 * removes a light from the scene.
 * if the second argument (del) is true, it deletes it as well.
 */
void Scene::RemoveLight(Light *light, bool del) {
	std::vector<Light*>::iterator iter = std::find(lights.begin(), lights.end(), light);
	
	if(iter != lights.end()) {
		lights.erase(iter);
		if(del) delete light;
	}
}

/**
 * removes a camera from the scene.
 * if the second argument (del) is true, it deletes it as well.
 */
void Scene::RemoveCamera(Camera *cam, bool del) {
	std::vector<Camera*>::iterator iter = std::find(cameras.begin(), cameras.end(), cam);
	
	if(iter != cameras.end()) {
		cameras.erase(iter);
		
		if(active_cam == cam) {
			if(cameras.size() > 0) {
				SetActiveCamera(cameras[0]);
			} else {
				SetActiveCamera(0);
			}
		}
		if(del) delete cam;
	}
}

/**
 * removes a curve from the scene.
 * if the second argument (del) is true, it deletes it as well.
 */
void Scene::RemoveCurve(Curve *curve, bool del) {
	std::vector<Curve*>::iterator iter = std::find(curves.begin(), curves.end(), curve);

	if(iter != curves.end()) {
		curves.erase(iter);
		if(del) delete curve;
	}
}


/** retrieves the pointer to an object by its name, or 0 if not found */
Object *Scene::GetObject(std::string name) {
	int size = objects.size();
	for(int i=0; i<size; i++) {
		if(objects[i]->GetName() == name) {
			return objects[i];
		}
	}
	return 0;
}

/** retrieves the pointer to a light by its name, or 0 if not found */
Light *Scene::GetLight(std::string name) {
	int size = lights.size();
	for(int i=0; i<size; i++) {
		if(lights[i]->GetName() == name) {
			return lights[i];
		}
	}
	return 0;
}

/** retrieves the pointer to a camera by its name, or 0 if not found */
Camera *Scene::GetCamera(std::string name) {
	int size = cameras.size();
	for(int i=0; i<size; i++) {
		if(cameras[i]->GetName() == name) {
			return cameras[i];
		}
	}
	return 0;
}

/** retrieves the pointer to a curve by its name, or 0 if not found */
Curve *Scene::GetCurve(std::string name) {
	int size = curves.size();
	for(int i=0; i<size; i++) {
		if(curves[i]->GetName() == name) {
			return curves[i];
		}
	}
	return 0;
}

/** returns the number of objects in the scene */
int Scene::GetObjectCount() const {
	return objects.size();
}

/** returns the number of lights in the scene */
int Scene::GetLightCount() const {
	return lights.size();
}

/** returns the number of cameras in the scene */
int Scene::GetCameraCount() const {
	return cameras.size();
}

/** returns the number of curves in the scene */
int Scene::GetCurveCounr() const {
	return curves.size();
}

/** set the active camera from which we shall render */
void Scene::SetActiveCamera(Camera *cam) {
	active_cam = cam;
	if(cam) UpdateCamera();
}

/** 
 * updates the state of the camera, recalculates ray-pixel
 * positions in world space (by calling CallPixelPositions()) etc.
 */
void Scene::UpdateCamera() {
	PRS prs = active_cam->GetPRS(anim.current_time);

	TargetCamera *tcam;
	if((tcam = dynamic_cast<TargetCamera*>(active_cam))) {
		// target camera
		Vector3 pos = prs.position;
		Vector3 targ = tcam->GetTarget();

		Vector3 n = (targ - pos).Normalized();
		Vector3 u = CrossProduct(active_cam->GetUpVector(), n).Normalized();
		Vector3 v = CrossProduct(n, u);

		scalar_t tx = DotProduct(u, pos);
		scalar_t ty = DotProduct(v, pos);
		scalar_t tz = DotProduct(n, pos);

		Matrix4x4 cam_trans, cam_rot;

		cam_trans.SetTranslation(Vector3(tx, ty, tz));
		cam_rot = Matrix4x4(	u.x, v.x, n.x, 0.0,
								u.y, v.y, n.y, 0.0,
								u.z, v.z, n.z, 0.0,
								0.0, 0.0, 0.0, 1.0);
		cam_xform = cam_rot * cam_trans;
	} else {
		// free camera
		cam_xform.SetTranslation(prs.position);
		cam_xform *= prs.rotation.GetRotationMatrix();
	}

	CalcPixelPositions();
}

const Animation &Scene::GetAnimation() const {
	return anim;
}

/** renders the scene */
void Scene::Render() const {
	if(!active_cam) return;

	ntimer timer;
	timer_reset(&timer);
	
	int i = 0;
	while(RenderScanline(i++));

	unsigned long rtime = timer_getmsec(&timer);
	info("rendering time: %.2f sec (%lu msec)", (float)rtime / 1000.0f, rtime);
}

/** renders a specified scanline (provided to aid paralelism) */
bool Scene::RenderScanline(int y) const {
	XRayContext *xrc = GetActiveXRayContext();
	if(!active_cam || y < 0 || y >= xrc->y) return false;

	scalar_t dist = (xrc->pix_pos[y * xrc->x + 1] - xrc->pix_pos[y * xrc->x]).Length();
	dist = dist / 2.0 + (dist - dist / 2.0) / 4.0;
	scalar_t dist2 = dist * 2.0;

	Ray ray;
	ray.origin = xrc->eye_pos;
	ray.energy = 1.0;
	ray.Enter(1.0);
	
	unsigned long *ptr = &xrc->pbuf->buffer[y * xrc->x];
	for(int x=0; x<xrc->x; x++) {
		Vector3 pix_pos = xrc->pix_pos[y * xrc->x + x];
		
		Vector3 col_acc;
		for(int i=0; i<xrc->rays_per_pixel; i++) {
			if(xrc->rays_per_pixel > 1) {
				Vector3 rnd_vec(frand(dist2) - dist, frand(dist2) - dist, frand(dist2) - dist);
				ray.dir = (pix_pos + rnd_vec) - xrc->eye_pos;
			} else {
				ray.dir = pix_pos - xrc->eye_pos;
			}
			ray.dir.Normalize();

			Color c = ShootRay(ray, 0);
			col_acc += Vector3(c.r, c.g, c.b);
		}

		*ptr++ = Color(col_acc / xrc->rays_per_pixel).GetPacked32();
	}

	if(ProgressCallback && (y % 4) == 0) {
		ProgressCallback((float)y / (float)xrc->y);
	}

	return true;
}
