/*
Copyright 2004 John Tsiombikas <nuclear@siggraph.org>

This file is part of the X-Ray ray-tracing library.

X-Ray is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

X-Ray is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with X-Ray; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @file scene.hpp */

#ifndef _SCENE_HPP_
#define _SCENE_HPP_

#include "object.hpp"
#include "animation.hpp"
#include "light.hpp"
#include "camera.hpp"
#include "gfx/curves.hpp"

namespace xray {

	class Scene {
	private:
		std::vector<Object*> objects;
		std::vector<Light*> lights;
		std::vector<Camera*> cameras;
		std::vector<Curve*> curves;

		Color background, ambient;

		Camera *active_cam;
		Matrix4x4 cam_xform;

		Animation anim;

		bool manage_elements;
		void (*ProgressCallback)(float);

		void CalcPixelPositions();

	public:
		Scene();
		~Scene();

		void SetProgressCallback(void (*ProgressCallback)(float));

		void AddObject(Object *obj);
		void AddLight(Light *light);
		void AddCamera(Camera *cam);
		void AddCurve(Curve *curve);

		void RemoveObject(Object *obj, bool del = false);
		void RemoveLight(Light *light, bool del = false);
		void RemoveCamera(Camera *cam, bool del = false);
		void RemoveCurve(Curve *curve, bool del = false);

		Object *GetObject(std::string name);
		Light *GetLight(std::string name);
		Camera *GetCamera(std::string name);
		Curve *GetCurve(std::string name);

		int GetObjectCount() const;
		int GetLightCount() const;
		int GetCameraCount() const;
		int GetCurveCounr() const;

		void SetActiveCamera(Camera *cam);
		void UpdateCamera();

		const Animation &GetAnimation() const;

		void Render() const;
		bool RenderScanline(int y) const;

		// stupid microsoft compiler...
#ifdef _MSC_VER
		friend Color ShootRay(const Ray &ray, int ray_depth);
		friend scalar_t ShadowRay(const Vector3 &pt, const Light *light);
		friend Color Shade(const Ray &ray, SurfPoint pt, const Object *obj, int ray_depth);
		friend Color ShadeMonteCarlo(const Ray &ray, SurfPoint pt, const Object *obj, int ray_depth);
#else
		friend Color xray::ShootRay(const Ray &ray, int ray_depth);
		friend scalar_t xray::ShadowRay(const Vector3 &pt, const Light *light);
		friend Color xray::Shade(const Ray &ray, SurfPoint pt, const Object *obj, int ray_depth);
		friend Color xray::ShadeMonteCarlo(const Ray &ray, SurfPoint pt, const Object *obj, int ray_depth);
#endif	// _MSC_VER
	};
}
		

#endif	// _SCENE_HPP_
